/*
    *
    * This file is a part of CoreStuff.
    * An activity viewer for CuboCore Application Suite
    * Copyright 2019 CuboCore Group
    *

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 3 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, write to the Free Software
    * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
    * MA 02110-1301, USA.
    *
*/

#include <QDebug>
#include <QScreen>
#include <QPainter>
#include <QFileInfo>
#include <QResizeEvent>

#include "global.h"

#include "pageactivites.h"
#include "pagehome.h"
#include "pagepins.h"
#include "pagereminders.h"
#include "pagesessions.h"
#include "pagetasks.h"
#include "lxqttray.h"

#include "slidingstackedwidget.h"
#include "corestuff.h"
#include "ui_corestuff.h"

#include <cprime/cprime.h>


corestuff::corestuff( QWidget *parent ):QWidget( parent ),ui( new Ui::corestuff )
{
	ui->setupUi( this );

	QPalette pltt( palette() );
	pltt.setColor( QPalette::Base, Qt::transparent );
	setPalette( pltt );

	loadSettings();
	startSetup();

	// Set coreapps page as active page
    on_home_clicked();

	ui->menu->setVisible( 0 );
	ui->addSession->setVisible( 0 );
	ui->deleteSession->setVisible( 0 );
	ui->editSessoion->setVisible( 0 );
	ui->clearActivites->setVisible( 0 );
}

corestuff::~corestuff()
{
	delete ui;
}

void corestuff::closeEvent(QCloseEvent *event)
{
    event->ignore();
    for (int i = 0; i < ui->pages->count(); i++) {
        if (ui->pages->widget(i)->children().count() == 2)
            if (ui->pages->widget(i)->children().at(1))
                ui->pages->widget(i)->children().at(1)->deleteLater();
    }
    event->accept();
}

void corestuff::startSetup()
{
    ui->user->setText(qgetenv("USER"));

	// set window size
    if ( touch ) {
        setWindowFlags(Qt::Window | Qt::FramelessWindowHint);
		setWindowState( Qt::WindowMaximized );
    } else if (desktop) {
        setFixedSize( qApp->primaryScreen()->size() );
        setWindowState( Qt::WindowFullScreen );
        setAttribute( Qt::WA_X11NetWmWindowTypeDesktop );
        bgImage = QImage( QString( sm->value( "CoreStuff", "Background" ) ) ).scaled( qApp->primaryScreen()->size(), Qt::KeepAspectRatioByExpanding, Qt::SmoothTransformation );
    }
    else {
		resize( 800,500 );
	}

	fswStart = new QFileSystemWatcher( this );
	connect( fswStart, &QFileSystemWatcher::fileChanged, [ = ]( const QString & path ) {
		reload( path );
	} );

    fswStart->addPaths( QStringList() << CPrime::Variables::CC_CoreApps_RecentActFilePath()
                        << CPrime::Variables::CC_CoreApps_PinsFilePath()
                        << CPrime::Variables::CC_CoreApps_SessionFilePath());

    // --------------------------------------------------------------

    if(desktop){
        LXQtTray *tray = new LXQtTray( this );
        connect( tray, SIGNAL( traySizeChanged( int ) ), this, SLOT( resizeDock( int ) ) );

        tray->setFixedHeight( 30 );
        tray->setIconSize( QSize( 24, 24 ) );

        ui->a->addWidget( tray );

        tray->resize( 0, 0 );
        tray->updateGeometry();
    }

    if(!showTasks){
        ui->tasks->setVisible(false);
    }
}

void corestuff::reload( const QString &path )
{
    QFileInfo fi( path );

    if ( fi.fileName() == CPrime::Variables::CC_CoreApps_RecentActFile() ) {
        if (showRecent) {
            if (pActivities) {
                pActivities->reload();
            }
        }
    }

    else if ( fi.fileName() == CPrime::Variables::CC_CoreApps_PinsFile() ) {
        if (pPins) {
            pPins->reload();
        }
    }

    else if ( fi.fileName() == CPrime::Variables::CC_CoreApps_SessionFile() ) {
        if (pSessions) {
            pSessions->reload();
        }
    }
}

/**
 * @brief Loads application settings
 */
void corestuff::loadSettings()
{
    touch = sm->value("CoreApps", "TouchMode");
    showTasks  = sm->value("CoreStuff", "TaskView");
    desktop = sm->value("CoreStuff", "Desktop");
    showRecent = sm->value("CoreApps", "ShowRecent");
}

void corestuff::pageClick( QPushButton *btn, int i, QString windowTitle )
{
	// all button checked false
	for ( QPushButton *b : ui->pageButtons->findChildren<QPushButton *>() ) {
		b->setChecked( false );
	}

	btn->setChecked( true );
	ui->pages->slideInIdx( i );
	currentPage = i;
    on_menu_clicked();
	setWindowTitle( windowTitle + " - CoreStuff" );
}

void corestuff::on_home_clicked()
{
    if (ui->pages->widget(0)->children().count() < 2) {
        pHome = new pagehome(this);
        ui->pHomeLayout->addWidget(pHome);
    }

    pageClick( ui->home, 0, "Home" );
    ui->menu->setVisible(false);
}


void corestuff::on_tasks_clicked()
{
    if (ui->pages->widget(1)->children().count() < 2) {
        pTasks = new pagetasks(this);
        pTasks->setCoreStuff(this);
        ui->pTasksLayout->addWidget(pTasks);
    }

    pageClick( ui->tasks, 1, "Tasks" );
    ui->menu->setVisible(false);
}

void corestuff::on_pins_clicked()
{
    if (ui->pages->widget(2)->children().count() < 2) {
        pPins = new pagepins(this);
        ui->pPinsLayout->addWidget(pPins);
    }

    pageClick( ui->pins, 2, "Pins" );
    ui->menu->setVisible(false);
}


void corestuff::on_activites_clicked()
{
    if (ui->pages->widget(3)->children().count() < 2) {
        pActivities = new pageactivites(this);
        ui->pActivitesLayout->addWidget(pActivities);

        // Configure Recent Activity
            if (showRecent) {
                QString raFile = CPrime::Variables::CC_CoreApps_RecentActFilePath();
                QFile file( raFile );

                if ( !file.exists() ) {
                    // You can get error
                    // Need a check here
                    file.open( QIODevice::ReadWrite | QIODevice::Text );
                    file.close();
                }
            }
    }

    pageClick( ui->activites, 3, "Activites" );
    ui->menu->setVisible(true);
}

void corestuff::on_session_clicked()
{
    if (ui->pages->widget(4)->children().count() < 2) {
        pSessions = new pagesessions(this);
        QList<QToolButton*> list;
        list.append(ui->menu);
        list.append(ui->addSession);
        list.append(ui->editSessoion);
        list.append(ui->deleteSession);
        pSessions->addControls(list);
        ui->pSessionsLayout->addWidget(pSessions);
    }

    pageClick( ui->session, 4, "Sessions" );
    ui->menu->setVisible(true);
}

void corestuff::on_reminders_clicked()
{
    if (ui->pages->widget(5)->children().count() < 2) {
        pReminders = new pagereminders(this);
        ui->pRemindersLayout->addWidget(pReminders);
    }

    pageClick(ui->reminders, 5, "Reminders");
    ui->menu->setVisible(false);
}

void corestuff::on_clearActivites_clicked()
{
    // link to pageactivites.cpp
    pActivities->clearActivities();
}

void corestuff::on_editSessoion_clicked()
{
    //link to pagesession.cpp
    pSessions->editSessoion();
}

void corestuff::on_deleteSession_clicked()
{
    //link to pagesession.cpp
    pSessions->deleteSession();
}

void corestuff::on_addSession_clicked()
{
    //link to pagesession.cpp
    pSessions->addSession();
}

void corestuff::on_menu_clicked()
{
    if (!ui->menu->isChecked()) {
        return;
    }

    if ( currentPage == 3 ) { // activites
        ui->addSession->setVisible( 0 );
        ui->deleteSession->setVisible( 0 );
        ui->editSessoion->setVisible( 0 );
        ui->clearActivites->setVisible( 1 );
    }

    else if ( currentPage == 4 ) { // session
        ui->addSession->setVisible( 1 );
        ui->deleteSession->setVisible( 0 );
        ui->editSessoion->setVisible( 0 );
        ui->clearActivites->setVisible( 0 );
    }

    else {
        ui->addSession->setVisible( 0 );
        ui->deleteSession->setVisible( 0 );
        ui->editSessoion->setVisible( 0 );
        ui->clearActivites->setVisible( 0 );
    }
}

void corestuff::handleMessages( const QString msg ) {

	// This is basically to reload the background
	if ( msg == "reload" ) {
		bgImage = QImage( sm->value( "CoreStuff", "Background" ) ).scaled( bgImage.size(), Qt::KeepAspectRatioByExpanding, Qt::SmoothTransformation );
		repaint();
	}
}

void corestuff::paintEvent( QPaintEvent *pEvent ) {

	QPainter painter( this );
	painter.drawImage( QRect( QPoint( 0, 0 ), bgImage.size() ), bgImage );

	painter.end();

	QWidget::paintEvent( pEvent );

	pEvent->accept();
};
