/*
    *
    * This file is a part of CoreStuff.
    * An activity viewer for CuboCore Application Suite
    * Copyright 2019 CuboCore Group
    *

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 3 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, write to the Free Software
    * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
    * MA 02110-1301, USA.
    *
*/


#include <QPainter>
#include <QListWidget>
#include <QDebug>
#include <QScreen>
#include <QCloseEvent>
#include <QTimer>
#include <QPluginLoader>
#include <QFuture>
#include <QGraphicsDropShadowEffect>
#include <QtConcurrent/QtConcurrent>

#include <cprime/appopenfunc.h>
#include <cprime/cplugininterface.h>
#include <cprime/cprime.h>
#include <cprime/applicationdialog.h>
#include <cprime/desktopfile.h>

#include "pagehome.h"
#include "ui_pagehome.h"


pagehome::pagehome(QWidget *parent) :QWidget(parent),ui(new Ui::pagehome)
{
    ui->setupUi(this);

    loadSettings();
    startSetup();
    setupWidgetView();
    setupAppsView();
}

pagehome::~pagehome()
{
    // Saving settings
    qDebug() << "WidgetViewWidth " << ui->splitter->sizes()[0];
    sm->setValue("CoreBox", "WidgetViewWidth", ui->splitter->sizes()[0]);
    sm->reload();

    delete ui;
}

/**
 * @brief Loads application settings
 */
void pagehome::loadSettings()
{
    widgetViewWidth = sm->value("CoreStuff", "WidgetViewWidth");
    iconViewIconSize = sm->value("CoreApps", "IconViewIconSize");
    touch = sm->value("CoreApps", "TouchMode");
    favouriteApps = sm->value("CoreApps", "Favourite");
    selectedPlugins = sm->value("CoreAction", "SelectedPlugins");
}

void pagehome::startSetup()
{
    ui->appsView->setIconSize( iconViewIconSize );
    ui->appsView->setGridSize( QSize( iconViewIconSize.width() * 3, iconViewIconSize.height() * 2 ) );

    if ( touch ) {
        connect( ui->appsView, &QListWidget::itemClicked,this,&pagehome::openSelectedApp);
    } else {
        connect( ui->appsView,&QListWidget::itemDoubleClicked,this,&pagehome::openSelectedApp);
    }

    QPalette pltt( palette() );
    QColor txtColor( pltt.color( QPalette::Text ) );

	QGraphicsDropShadowEffect *effect = new QGraphicsDropShadowEffect(this);
	effect->setBlurRadius( 0 );
	effect->setColor( QColor( 0xff - txtColor.red(), 0xff - txtColor.green(), 0xff - txtColor.blue() ) );
	effect->setOffset( 1, 1 );

	ui->appsView->setGraphicsEffect(effect);
}

void pagehome::setupWidgetView()
{
    int coreactionWidth = widgetViewWidth; //static_cast<int>(width * .15);

    int totalWidth = this->size().width();
    int leftWidth = totalWidth - coreactionWidth;

    QList<int> list = {coreactionWidth, leftWidth};
    ui->splitter->setSizes(list);

    QTimer *timer = new QTimer( this );
    connect( timer, &QTimer::timeout, this, &pagehome::updateTime );
    timer->start( 1000 );

    /* All Plugins */
    QStringList pluginPaths = selectedPlugins;
    if ( not pluginPaths.count() ) {
        CPrime::InfoFunc::messageEngine(
            "Plugins not found. Please install plugins and ensure they are enabled in CoreGarage.",
            CPrime::MessageType::Info,
            this
        );
    }

    Q_FOREACH( QString pPath, pluginPaths ) {

        QPluginLoader loader(pPath);
        QObject *pObject = loader.instance();
        if (pObject) {
            plugininterface *plugin = qobject_cast<plugininterface*>(pObject);
            if ( !plugin ) {
                qWarning() << "Plugin Error:" << loader.errorString();
                continue;
            }

            QWidget *w = plugin->widget( this );
            ui->widgetLayout->insertWidget( 0, w );
        } else {
            qWarning() << "Plugin Error:" << loader.errorString();
        }
    }
}

void pagehome::setupAppsView()
{
    QStringList appList = favouriteApps;
    foreach ( QString app, appList ) {
        DesktopFile file = DesktopFile( CPrime::Variables::CC_System_AppPath() + "/" + app + ".desktop" );
        QIcon icon = ApplicationDialog::searchAppIcon( file );
        icon = CPrime::ThemeFunc::resizeIcon( icon, ui->appsView->iconSize() );
        QListWidgetItem *item = new QListWidgetItem( icon, file.getName() );
        item->setData( Qt::UserRole, file.getFileName() );
        item->setSizeHint( QSize( 110, 100 ) );
        ui->appsView->addItem( item );
    }

    QList<DesktopFile> list = ApplicationDialog::getApplications();

    std::sort( list.begin(), list.end(), [ ]( const DesktopFile& lhs, const DesktopFile& rhs ) {
        return lhs.getName() < rhs.getName();
    } );

    QFuture<void> f = QtConcurrent::run( [this, list, appList] {
        foreach ( DesktopFile df, list ) {
            if ( appList.contains( df.getPureFileName() ) ) {
                continue;
            }

            QIcon icon = ApplicationDialog::searchAppIcon( df );
            icon = CPrime::ThemeFunc::resizeIcon( icon, ui->appsView->iconSize() );

            QListWidgetItem *item = new QListWidgetItem( icon, df.getName() );
            item->setData( Qt::UserRole, df.getFileName() );
            item->setSizeHint( QSize( 110, 100 ) );
            ui->appsView->addItem( item );
        }
    } );

    QFutureWatcher<void> *fw = new QFutureWatcher<void>();
    fw->setFuture( f );
    connect( fw, &QFutureWatcher<void>::finished, [&]() {
        ui->appsView->setSizeAdjustPolicy( QListWidget::AdjustToContents );
    } );
}

void pagehome::updateTime()
{
    QDateTime dt = QDateTime::currentDateTime();
    ui->day->setText( dt.toString( "hh:mm AP\ndddd, MMM dd" ) );
}

void pagehome::openSelectedApp( QListWidgetItem *item )  // open SpeedDial on CoreApps
{
    // Load desktop file for application
    DesktopFile df = DesktopFile( item->data( Qt::UserRole ).toString() );
    QString exec = df.getExec();
    QStringList temp = exec.split( ' ' );
    if ( temp.count() ) {
        QString app = temp.at( 0 );
        if ( app.count() ) {
            exec = exec.remove( "%u", Qt::CaseInsensitive ).remove( "%f", Qt::CaseInsensitive );
            if ( exec.at( exec.count() - 1 ) == ' ' )
                exec = exec.remove( exec.count() - 1, 1 );
            CPrime::AppOpenFunc::systemAppOpener( exec, QStringList() );
        } else {
            qDebug() << "func( on_appCollect_itemDoubleClicked ) : App name empty.";
            return;
        }
    } else {
        qDebug() << "func( on_appCollect_itemDoubleClicked ) : Exec command empty.";
        return;
    }
}
