/*
    *
    * This file is a part of CoreStuff.
    * An activity viewer for CuboCore Application Suite
    * Copyright 2019 CuboCore Group
    *

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 3 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, write to the Free Software
    * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
    * MA 02110-1301, USA.
    *
*/

#include "corestuff.h"
#include "pagetasks.h"
#include "ui_pagetasks.h"


pagetasks::pagetasks(QWidget *parent) : QWidget(parent), ui(new Ui::pagetasks)
{
    ui->setupUi(this);

    loadSettings();
    startSetup();
    setupTaskView();
}

pagetasks::~pagetasks()
{
    delete ui;
}

/**
 * @brief Loads application settings
 */
void pagetasks::loadSettings()
{
    listViewIconSize = sm->value("CoreApps", "ListViewIconSize");
    touch = sm->value("CoreApps", "TouchMode");
}

void pagetasks::startSetup()
{
    ui->taskView->setIconSize(listViewIconSize);
}

void pagetasks::setupTaskView()
{
    // Initializing...
    m_x11support = new X11Support();
    cXCV.setX11Support(m_x11support);
    qApp->installNativeEventFilter(&cXCV);
    installEventFilter(m_x11support);

    connect(ui->taskView, &QListWidget::itemClicked, this, &pagetasks::setFocusApp);
    connect(ui->cl, SIGNAL(clicked()), this, SLOT(closeActiveApp()));

    // Register for notifications about window property changes.
    connect(X11Support::instance(), SIGNAL(windowPropertyChanged(ulong, ulong)), this, SLOT(windowPropertyChanged(ulong, ulong)));
    connect(X11Support::instance(), SIGNAL(windowReconfigured(ulong, int, int, int, int)), this, SLOT(windowReconfigured(ulong, int, int, int, int)));

    updateClientList();
    updateActiveWindow();
}


void pagetasks::closeActiveApp()
{
    if (ui->taskView->currentItem()) {
        X11Support::closeWindow(ui->taskView->currentItem()->data(Qt::UserRole).value<Info>().windowId);
    }
}

void pagetasks::updateClientList()
{
    ui->taskView->clear();
    QVector<unsigned long> windows = X11Support::getWindowPropertyWindowsArray(
                                         X11Support::rootWindow(), "_NET_CLIENT_LIST");

    // Handle new clients.
    for (int i = 0; i < windows.size(); i++) {
        if (!m_clients.contains(windows[i])) {
            // Skip our own windows.
            if (QWidget::find(windows[i]) != nullptr) {
                continue;
            }

            bool m_visible;
            QVector<unsigned long> windowTypes = X11Support::getWindowPropertyAtomsArray(windows[i], "_NET_WM_WINDOW_TYPE");
            QVector<unsigned long> windowStates = X11Support::getWindowPropertyAtomsArray(windows[i], "_NET_WM_STATE");

            // Show only regular windows in dock.
            // When no window type is set, assume it's normal window.
            m_visible = (windowTypes.size() == 0) || (windowTypes.size() == 1 && windowTypes[0] == X11Support::atom("_NET_WM_WINDOW_TYPE_NORMAL"));

            // Don't show window if requested explicitly in window states.
            if (windowStates.contains(X11Support::atom("_NET_WM_STATE_SKIP_TASKBAR"))) {
                m_visible = false;
            }

            if (m_visible) {
                // Do nothing
            }

            if (!m_visible) {
                continue;
            }

            m_clients[windows[i]] = new ClientT(windows[i]);

            QListWidgetItem *item = new QListWidgetItem();

            Info info;
            info.windowId = windows[i];
            info.m_minimized = X11Support::getWindowMinimizedState(windows[i]);

            item->setText(m_clients[windows[i]]->name());

            item->setData(Qt::UserRole, QVariant::fromValue(info));
            item->setIcon(X11Support::getWindowIcon(windows[i]));
            ui->taskView->addItem(item);
        }
    }

    m_in_loop.clear();

    // Handle removed clients.
    for (;;) {
        bool clientRemoved = false;

        foreach (ClientT *client, m_clients) {

            if (m_in_loop.contains(client)) {
                clientRemoved = false;
                break;
            }

            int handle = static_cast<int>(client->handle());

            delete m_clients[static_cast<unsigned long>(handle)];
            m_clients.remove(static_cast<unsigned long>(handle));
            clientRemoved = true;
            break;

            if (!windows.contains(static_cast<unsigned long>(handle))) {
                if (!m_in_loop.contains(client)) {
                    m_in_loop.append(client);
                }

                delete m_clients[static_cast<unsigned long>(handle)];
                m_clients.remove(static_cast<unsigned long>(handle));
                clientRemoved = true;
                break;
            }
        }

        if (!clientRemoved) {
            break;
        }
    }
}

void pagetasks::updateActiveWindow()
{
    unsigned long activeWindow = X11Support::getWindowPropertyWindow(
                                     X11Support::rootWindow(), "_NET_ACTIVE_WINDOW");

    if (activeWindow) {
        m_activeWindow = activeWindow;
    }
}

void pagetasks::windowPropertyChanged(unsigned long window, unsigned long atom)
{
    if (window == X11Support::rootWindow()) {
        if (atom == X11Support::atom("_NET_CLIENT_LIST")) {
            updateClientList();
        }

        if (atom == X11Support::atom("_NET_ACTIVE_WINDOW")) {
            updateActiveWindow();
        }

        return;
    }

    if (m_clients.contains(window)) {
        m_clients[window]->windowPropertyChanged(atom);
    }
}

void pagetasks::windowReconfigured(unsigned long window, int x, int y, int width, int height)
{
    Q_UNUSED(window)
    Q_UNUSED(x)
    Q_UNUSED(y)
    Q_UNUSED(width)
    Q_UNUSED(height)

    if (window == X11Support::rootWindow()) {
        return;
    }

    updateClientList();
}

void pagetasks::setFocusApp(QListWidgetItem *item)
{
    if (item) {
        bool isMinimized = item->data(Qt::UserRole).value<Info>().m_minimized;
        unsigned long windowId = item->data(Qt::UserRole).value<Info>().windowId;

        if (isMinimized) {
            X11Support::activateWindow(windowId);
            Info inf;
            inf.m_minimized = false;
            inf.windowId = windowId;
            item->setData(Qt::UserRole, QVariant::fromValue(inf));
            if (m_cstuff) {
                m_cstuff->hide();
            }
        } else if (!isMinimized && windowId == m_activeWindow) {
            // qDebug() << "Setting to minimize";
//            X11Support::minimizeWindow(windowId);
//            Info inf;
//            inf.m_minimized = true;
//            inf.windowId = windowId;
//            item->setData(Qt::UserRole, QVariant::fromValue(inf));
        } else if (!isMinimized && windowId != m_activeWindow) {
            X11Support::activateWindow(windowId);
            Info inf;
            inf.m_minimized = false;
            inf.windowId = windowId;
            item->setData(Qt::UserRole, QVariant::fromValue(inf));
            if (m_cstuff) {
                m_cstuff->hide();
            }
        }
    }

    if (ui->taskView->currentItem()) {
        ui->cl->setVisible(1);
    } else {
        ui->cl->setVisible(0);
    }

}
